from pypika import AliasedQuery, Criterion

from metrics_layer.core.sql.query_base import MetricsLayerQueryBase
from metrics_layer.core.model.filter import LiteralValueCriterion
from metrics_layer.core.sql.query_dialect import query_lookup


class MetricsLayerMergedResultsQuery(MetricsLayerQueryBase):
    """ """

    def __init__(self, definition: dict) -> None:
        self.query_lookup = query_lookup
        super().__init__(definition)

    def get_query(self, semicolon: bool = True):
        # Build the base_cte table from the referenced queries + join them with all dimensions
        base_cte_query = self.build_cte_from()

        # Add all columns in the SELECT clause
        select = self.get_select_columns()
        complete_query = base_cte_query.select(*select)

        if self.having:
            where = self.get_where_from_having(project=self.project)
            complete_query = complete_query.where(Criterion.all(where))

        sql = str(complete_query.limit(self.limit))
        if semicolon:
            sql += ";"
        return sql

    def build_cte_from(self):
        base_cte_query = self._base_query()
        for join_hash, query in self.queries_to_join.items():
            base_cte_query = base_cte_query.with_(query, join_hash)

        for i, join_hash in enumerate(self.join_hashes):
            if i == 0:
                base_cte_query = base_cte_query.from_(AliasedQuery(join_hash))
            else:
                criteria = self._build_join_criteria(self.join_hashes[0], join_hash)
                base_cte_query = base_cte_query.outer_join(AliasedQuery(join_hash)).on(criteria)

        return base_cte_query

    def _build_join_criteria(self, first_query_alias, second_query_alias):
        no_dimensions = all(len(v) == 0 for v in self.query_dimensions.values())
        # No dimensions to join on, the query results must be just one number each
        if no_dimensions:
            return LiteralValueCriterion("1=1")

        join_criteria = []
        for i in range(len(self.query_dimensions[first_query_alias])):
            first_field = self.query_dimensions[first_query_alias][i]
            second_field = self.query_dimensions[second_query_alias][i]
            first_alias_and_id = f"{first_query_alias}.{first_field.alias(with_view=True)}"
            second_alias_and_id = f"{second_query_alias}.{second_field.alias(with_view=True)}"
            join_criteria.append(f"{first_alias_and_id}={second_alias_and_id}")

        return LiteralValueCriterion(" and ".join(join_criteria))

    # Code to handle SELECT portion of query
    def get_select_columns(self):
        select = []
        existing_aliases = []
        for join_hash, field_set in sorted(self.query_metrics.items()):
            for field in field_set:
                alias = field.alias(with_view=True)
                if alias not in existing_aliases:
                    select.append(self.sql(f"{join_hash}.{alias}", alias=alias))
                    existing_aliases.append(alias)

        dimension_sql = {}
        for join_hash, field_set in sorted(self.query_dimensions.items()):
            for field in field_set:
                alias = field.alias(with_view=True)
                if alias not in dimension_sql:
                    dimension_sql[alias] = f"{join_hash}.{alias}"
                else:
                    dimension_sql[alias] = f"ifnull({dimension_sql[alias]}, {join_hash}.{alias})"

        for alias, sql in dimension_sql.items():
            select.append(self.sql(sql, alias=alias))
            existing_aliases.append(alias)

        for field in self.merged_metrics:
            alias = field.alias(with_view=True)
            if alias not in existing_aliases:
                select.append(self.sql(field.strict_replaced_query(), alias=alias))
                existing_aliases.append(alias)

        return select



from pypika import Query
from pypika.dialects import (
    PostgreSQLQueryBuilder,
    RedShiftQueryBuilder,
    SnowflakeQueryBuilder,
)
from pypika.enums import Dialects

from metrics_layer.core.model.definitions import Definitions

SnowflakeQueryBuilder.ALIAS_QUOTE_CHAR = None
RedShiftQueryBuilder.ALIAS_QUOTE_CHAR = None
RedShiftQueryBuilder.QUOTE_CHAR = None
PostgreSQLQueryBuilder.ALIAS_QUOTE_CHAR = None
PostgreSQLQueryBuilder.QUOTE_CHAR = None


class SnowflakeQuery(Query):
    """
    Defines a query class for use with Snowflake.
    """

    @classmethod
    def _builder(cls, **kwargs) -> SnowflakeQueryBuilder:
        return SnowflakeQueryBuilder(**kwargs)


class PostgresQuery(Query):
    """
    Defines a query class for use with Snowflake.
    """

    @classmethod
    def _builder(cls, **kwargs) -> PostgreSQLQueryBuilder:
        return PostgreSQLQueryBuilder(**kwargs)


class RedshiftQuery(Query):
    """
    Defines a query class for use with Amazon Redshift.
    """

    @classmethod
    def _builder(cls, **kwargs) -> "RedShiftQueryBuilder":
        return RedShiftQueryBuilder(dialect=Dialects.REDSHIFT, **kwargs)


class BigQueryQuery(Query):
    """
    Defines a query class for use with BigQuery.
    """

    @classmethod
    def _builder(cls, **kwargs) -> SnowflakeQueryBuilder:
        return SnowflakeQueryBuilder(**kwargs)


query_lookup = {
    Definitions.snowflake: SnowflakeQuery,
    Definitions.bigquery: BigQueryQuery,
    Definitions.redshift: RedshiftQuery,
    Definitions.postgres: PostgresQuery,
}



# This takes an incoming SQL statement and converts the MQL() section of it
# into our internal format, asks the main query method to resolve the sql
# for that configuration then replaces the MQL() section of the original string
# with the correct SQL

from copy import deepcopy

import sqlparse
from sqlparse.sql import Function, Identifier, IdentifierList, Statement, Where
from sqlparse.tokens import Keyword, Whitespace

from metrics_layer.core.sql.query_errors import ParseError
from metrics_layer.core.sql.resolve import SQLQueryResolver


class MQLConverter:
    """
    Syntax here is:

    MQL(
        metric_name
        BY
        dimension
        WHERE
        condition
        HAVING
        having_condition
        ORDER BY
        metric_name
    )

    which will resolve to the SQL query that gives that result.

    The MQL feature can be used to compose SQL queries as well, as follows

    SELECT
        countries.country,
        metric_by_country.metric_name
    SELECT countries
        LEFT JOIN
            MQL(
                metric_name
                BY
                country
                WHERE
                condition
                HAVING
                having_condition
                ORDER BY
                metric_name
            ) as metric_by_country
            ON metric_by_country.country=countries.country

    The syntax can also be used for event / funnel queries, as follows:

    SELECT
        *
    FROM MQL(
            total_revenue, number_of_users

            FOR events
            FUNNEL event_name = 'user_created'
            THEN event_name = 'complete_onboarding' as onboarding,
                        event_name = 'purchase' as made_a_purchase
            WITHIN 3 days
            BY region, new_vs_repeat
            WHERE region != 'West' AND new_vs_repeat <> 'New'
        ) as subquery
    """

    def __init__(self, sql: str, project, **kwargs):
        self._function_name = "MQL"
        self.sql = sql
        self.project = project
        self.kwargs = kwargs

    def get_query(self):
        converted_sql = deepcopy(self.sql)
        statement = sqlparse.parse(self.sql)[0]
        for token in statement:
            # Handle case without an alias
            if isinstance(token, Function):
                mql_as_sql = self.parse_and_resolve_mql(token)
                if mql_as_sql:
                    # Do the replace for the MQL() function with the resolved SQL
                    converted_sql = converted_sql.replace(str(token), mql_as_sql)

            # Handle error with parenthesis
            if isinstance(token, Identifier) and str(token).upper() == self._function_name:
                raise ParseError(
                    f"Expected beginning and ending parenthesis around statement beginning with "
                    f"{str(token)} in statement {self.sql}"
                )

            # Handle case with alias
            if isinstance(token, Identifier) and f"{self._function_name}(" in str(token).upper():
                # We have to index into the first token here because the MQL is the first part of the token
                # with an alias e.g. [part 1] MQL(blah by blah) [part 2] as [part 3] alias
                mql_token = token[0]
                mql_as_sql = self.parse_and_resolve_mql(mql_token)
                if mql_as_sql:
                    # Do the replace for the MQL() function with the resolved SQL
                    converted_sql = converted_sql.replace(str(mql_token), mql_as_sql)

        if converted_sql[-1] != ";":
            converted_sql += ";"
        return converted_sql

    def parse_and_resolve_mql(self, token):
        function_name = token[0]
        if str(function_name).upper() == self._function_name:
            mql_as_sql = self.resolve_mql_statement(token[1])
            wrap_parenthesis = str(token) != self.sql
            if wrap_parenthesis:
                mql_as_sql = f"({mql_as_sql})"
            return mql_as_sql
        return

    def resolve_mql_statement(self, mql_statement):
        metrics, dimensions, where, having, order_by = [], [], [], [], []
        funnel_steps, funnel_for, funnel_within = [], None, {}
        mode = "metrics"
        # We do this to trim off the starting and ending parenthesis
        for token in mql_statement[1:-1]:
            if mode == "for" and token.ttype != Whitespace:
                funnel_for = str(token.value).split(" ")[0].lower().strip()
                mode = "funnel"
                funnel_step = []
                continue

            is_conjunction = str(token).strip().upper() in {"AND", "OR"}
            if (token.ttype == Keyword or self._is_funnel(token)) and not is_conjunction:
                mode = self._resolve_mode(token, mode)
                if mode == "funnel" and str(token).upper() != "THEN":
                    funnel_step = []
                    continue
                elif mode == "dimensions" and str(token).upper() in {"SOURCE"}:
                    self._add_by_mode(metrics, dimensions, mode, str(token))

            if mode == "funnel":
                if str(token).upper() == "THEN":
                    funnel_steps.append(deepcopy(funnel_step))
                    funnel_step = []
                    continue
                elif "WITHIN" in str(token).upper():
                    lowered = str(token).lower()
                    funnel_step.append(str(token)[: lowered.index("within")].strip())
                    funnel_steps.append(deepcopy(funnel_step))
                    mode = "within"
                    continue
                else:
                    funnel_step.append(str(token))
                    continue

            if mode == "within" and isinstance(token, Identifier):
                n, unit = str(token).strip().split(" ")
                funnel_within = {"value": int(n), "unit": unit.lower()}
                mode = "metrics"
                continue

            if mode == "having":
                having.append(token)
                continue

            if mode == "order_by":
                order_by.append(token)
                continue

            if isinstance(token, Identifier) and not self._is_funnel(token):
                identifier = str(token)
                self._add_by_mode(metrics, dimensions, mode, identifier)

            if isinstance(token, IdentifierList):
                for id_token in token.get_identifiers():
                    identifier = str(id_token)
                    self._add_by_mode(metrics, dimensions, mode, identifier)

            if isinstance(token, Where):
                where = token

        if funnel_within != {} and len(funnel_steps) > 0:
            steps = ["".join(s).strip() for s in funnel_steps]
            funnel = {"steps": steps, "within": funnel_within}
            if funnel_for:
                funnel["view_name"] = funnel_for
        else:
            funnel = {}

        # For all of these we do [1:] to ignore the keyword at the beginning of the statement
        where_literal = self._tokens_to_sql(where[1:])
        having_literal = self._tokens_to_sql(having[1:])
        order_by_literal = self._tokens_to_sql(order_by[1:])

        resolver = SQLQueryResolver(
            metrics=metrics,
            dimensions=dimensions,
            funnel=funnel,
            where=where_literal,
            having=having_literal,
            order_by=order_by_literal,
            project=self.project,
            **self.kwargs,
        )
        return resolver.get_query(semicolon=False)

    def _resolve_mode(self, token, mode):
        keyword = str(token).upper()
        if keyword == "BY":
            mode = "dimensions"
        elif keyword == "HAVING":
            mode = "having"
        elif keyword == "ORDER BY":
            mode = "order_by"
        elif keyword == "FOR":
            mode = "for"
        elif keyword in {"FUNNEL", "THEN"}:
            mode = "funnel"
        elif keyword == "WITHIN":
            mode = "within"
        else:
            mode = mode
        return mode

    def _add_by_mode(self, metrics: list, dimensions: list, mode: str, identifier: str):
        if mode == "metrics":
            metrics.append(identifier)
        elif mode == "dimensions":
            dimensions.append(identifier)
        else:
            raise ParseError(f"Could not parse identifier {identifier} in statement {self.sql}")

    @staticmethod
    def _tokens_to_sql(tokens: list):
        if len(tokens) == 0:
            return
        return str(Statement(tokens)).strip()

    def _is_funnel(self, token):
        return str(token).upper().strip() == "FUNNEL"



import json
import hashlib
from metrics_layer.core.exceptions import QueryError
from collections import defaultdict
from copy import deepcopy

from metrics_layer.core.sql.query_merged_results import MetricsLayerMergedResultsQuery
from metrics_layer.core.sql.single_query_resolve import SingleSQLQueryResolver


class MergedSQLQueryResolver(SingleSQLQueryResolver):
    def __init__(
        self,
        metrics: list,
        dimensions: list = [],
        funnel: dict = {},
        where: str = None,  # Either a list of json or a string
        having: str = None,  # Either a list of json or a string
        order_by: str = None,  # Either a list of json or a string
        model=None,
        project=None,
        **kwargs,
    ):
        if funnel != {}:
            raise QueryError("Funnel queries are not supported in merged results queries")

        self.field_lookup = {}
        self.no_group_by = False
        self.has_cumulative_metric = False
        self.verbose = kwargs.get("verbose", False)
        self.select_raw_sql = kwargs.get("select_raw_sql", [])
        self.suppress_warnings = kwargs.get("suppress_warnings", False)
        self.limit = kwargs.get("limit")
        self.return_pypika_query = kwargs.get("return_pypika_query")
        self.force_group_by = kwargs.get("force_group_by", False)
        self.kwargs = kwargs
        self.project = project
        self.metrics = metrics
        self.dimensions = dimensions
        self.parse_field_names(where, having, order_by)
        self.model = model

    def get_query(self, semicolon: bool = True):
        self.parse_field_names(self.where, self.having, self.order_by)
        self.derive_sub_queries()

        queries_to_join = {}
        join_hashes = list(self.query_metrics.keys())
        for k in self.query_dimensions.keys():
            if k not in join_hashes:
                join_hashes.append(k)

        join_hash_readability_lookup = {
            j: f"{j.split('__')[0]}__cte_subquery_{i}" for i, j in enumerate(sorted(join_hashes))
        }
        print(join_hash_readability_lookup)

        readable_join_hashes = []
        for join_hash in join_hashes:
            metrics = [f.id() for f in self.query_metrics.get(join_hash, [])]
            dimensions = [f.id() for f in self.query_dimensions.get(join_hash, [])]

            # Overwrite the limit arg because these are subqueries
            kws = {**self.kwargs, "limit": None, "return_pypika_query": True}
            resolver = SingleSQLQueryResolver(
                metrics=metrics,
                dimensions=dimensions,
                where=self.query_where[join_hash],
                having=[],
                order_by=[],
                model=self.model,
                project=self.project,
                **kws,
            )
            query = resolver.get_query(semicolon=False)
            readable_hash = join_hash_readability_lookup[join_hash]
            readable_join_hashes.append(readable_hash)
            queries_to_join[readable_hash] = query

        readable_metrics = {join_hash_readability_lookup[k]: m for k, m in self.query_metrics.items()}
        readable_dimensions = {join_hash_readability_lookup[k]: d for k, d in self.query_dimensions.items()}
        query_config = {
            "merged_metrics": self.merged_metrics,
            "query_metrics": readable_metrics,
            "query_dimensions": readable_dimensions,
            "having": self.having,
            "queries_to_join": queries_to_join,
            "join_hashes": list(sorted(readable_join_hashes)),
            "query_type": resolver.query_type,
            "limit": self.limit,
            "project": self.project,
        }
        merged_result_query = MetricsLayerMergedResultsQuery(query_config)
        query = merged_result_query.get_query(semicolon=semicolon)

        return query

    def derive_sub_queries(self):
        self.query_metrics = defaultdict(list)
        self.merged_metrics = []
        self.secondary_metrics = []
        self.dimension_fields = []

        for metric in self.metrics:
            field = self.project.get_field(metric)
            if field.is_merged_result:
                self.merged_metrics.append(field)
            else:
                self.secondary_metrics.append(field)

        for dimension in self.dimensions:
            self.dimension_fields.append(self.project.get_field(dimension))

        for merged_metric in self.merged_metrics:
            for ref_field in merged_metric.referenced_fields(merged_metric.sql):
                if isinstance(ref_field, str):
                    raise QueryError(f"Unable to find the field {ref_field} in the project")
                if ref_field.canon_date is None:
                    raise QueryError(
                        f"Could not find a date field associated with metric {ref_field.name} "
                        "in the project. \n\nMake sure you have this defined either in the view "
                        "with the property 'default_date' or on the metric under 'canon_date'"
                    )
                join_group_hash = self._join_hash_key(ref_field)
                key = self._cte_name_from_parts(ref_field.canon_date, join_group_hash)
                self.query_metrics[key].append(ref_field)

        for field in self.secondary_metrics:
            join_group_hash = self._join_hash_key(field)
            if field.canon_date is None:
                raise QueryError(
                    f"Could not find a date field associated with metric {field.name} "
                    "in the project. \n\nMake sure you have this defined either in the view "
                    "with the property 'default_date' or on the metric under 'canon_date'"
                )
            key = self._cte_name_from_parts(field.canon_date, join_group_hash)
            if key in self.query_metrics:
                already_in_query = any(field.id() in f.id() for f in self.query_metrics[key])
                if not already_in_query:
                    self.query_metrics[key].append(field)
            else:
                self.query_metrics[key].append(field)

        self.query_metrics = self.deduplicate_fields(self.query_metrics)

        dimension_mapping, canon_dates, used_join_hashes = self._canon_date_mapping()

        mappings = self.model.get_mappings(dimensions_only=True)
        for key, map_to in mappings.items():
            for other_join_hash in used_join_hashes:
                if map_to["to_join_hash"] in other_join_hash:
                    map_to["from_join_hash"] = other_join_hash
                dimension_mapping[key].append(deepcopy(map_to))

        self.query_dimensions = defaultdict(list)
        for dimension in self.dimensions:
            field = self.project.get_field(dimension)
            field_key = f"{field.view.name}.{field.name}"
            join_group_hash = self._join_hash_key(field)
            if field_key in canon_dates:
                join_hash = f'{field_key.replace(".", "_")}__{join_group_hash}'
                self.query_dimensions[join_hash].append(field)

                dimension_group = field.dimension_group
                for mapping_info in dimension_mapping[field_key]:
                    key = f"{mapping_info['field']}_{dimension_group}"
                    ref_field = self.project.get_field(key)
                    self.query_dimensions[mapping_info["from_join_hash"]].append(ref_field)
            else:
                not_in_metrics = True
                for join_hash in used_join_hashes:
                    # If the dimension is available in the join subgraph as the metric, attach it
                    if any(jg in join_hash for jg in field.join_graphs()):
                        self.query_dimensions[join_hash].append(field)
                        not_in_metrics = False
                    else:
                        if field_key not in dimension_mapping:
                            raise QueryError(
                                f"Could not find mapping from field {field_key} to other views. "
                                "Please add a mapping to your model definition to allow the mapping "
                                "if you'd like to use this field in a merged result query."
                            )
                        for mapping_info in dimension_mapping[field_key]:
                            ref_field = self.project.get_field(mapping_info["field"])
                            if mapping_info["from_join_hash"] in self.query_metrics:
                                self.query_dimensions[mapping_info["from_join_hash"]].append(ref_field)
                            else:
                                canon_date = ref_field.canon_date.replace(".", "_")
                                existing_join_hashes = (
                                    join_hash
                                    for join_hash in used_join_hashes
                                    if mapping_info["to_join_hash"] in join_hash
                                )
                                join_hash = next(existing_join_hashes, None)
                                default_join_hash = f"{canon_date}__{mapping_info['to_join_hash']}"
                                key = join_hash if join_hash else default_join_hash
                                self.query_dimensions[key].append(ref_field)

                if not_in_metrics:
                    field_key = f"{field.view.name}.{field.name}"
                    canon_date = field.canon_date.replace(".", "_")
                    key = f"{canon_date}__{join_group_hash}"
                    self.query_metrics[key] = []
                    self.query_dimensions[key].append(field)
                    if field_key in dimension_mapping:
                        for mapping_info in dimension_mapping[field_key]:
                            ref_field = self.project.get_field(mapping_info["field"])
                            canon_date = ref_field.canon_date.replace(".", "_")
                            mapped_key = f"{canon_date}__{mapping_info['to_join_hash']}"
                            self.query_dimensions[mapped_key].append(ref_field)

        self.query_dimensions = self.deduplicate_fields(self.query_dimensions)

        keys = list(self.query_metrics.keys()) + list(self.query_dimensions.keys())
        unique_keys = sorted(list(set(keys)), key=lambda x: keys.index(x))
        self.query_where = defaultdict(list)
        for where in self.where:
            metric_canon_dates = {f.canon_date for v in self.query_metrics.values() for f in v}
            field = self.project.get_field(where["field"])
            is_canon_date = any(f"{field.view.name}.{field.name}" == d for d in metric_canon_dates)
            dimension_group = field.dimension_group
            join_group_hash = self._join_hash_key(field)
            added_filter = {join_hash: False for join_hash in unique_keys}
            for join_hash in unique_keys:
                join_hash_with_canon_date = f"{field.view.name}_{field.name}__{join_group_hash}"
                joinable_not_canon_date = not is_canon_date and join_group_hash in join_hash
                is_canon_date_same = is_canon_date and join_hash_with_canon_date in join_hash

                if joinable_not_canon_date or is_canon_date_same:
                    self.query_where[join_hash].append(where)
                    added_filter[join_hash] = True
                else:
                    key = f"{field.view.name}.{field.name}"
                    for mapping_info in dimension_mapping[key]:
                        if mapping_info["from_join_hash"] == join_hash:
                            if dimension_group:
                                key = f"{mapping_info['field']}_{dimension_group}"
                            else:
                                key = mapping_info["field"]
                            ref_field = self.project.get_field(key)
                            mapped_where = deepcopy(where)
                            mapped_where["field"] = ref_field.id()
                            self.query_where[join_hash].append(mapped_where)
                            added_filter[join_hash] = True
            # This handles the case where the where field is joined in and not in a mapping
            for join_hash in self.query_metrics.keys():
                if not added_filter[join_hash]:
                    self.query_where[join_hash].append(where)

        clean_wheres = defaultdict(list)
        for k, v in self.query_where.items():
            hashes, lookup = [], {}
            for item in v:
                h = self.hash_dict(item)
                lookup[h] = item
                hashes.append(h)

            sorted_hashes = sorted(list(set(hashes)), key=lambda x: hashes.index(x))
            clean_wheres[k] = [lookup[h] for h in sorted_hashes]
        self.query_where = clean_wheres

    def _canon_date_mapping(self):
        canon_dates, join_hashes = [], []
        dimension_mapping = defaultdict(list)
        for merged_metric in self.merged_metrics:
            for ref_field in merged_metric.referenced_fields(merged_metric.sql):
                canon_dates.append(ref_field.canon_date)

        joinable_sets = []
        for field in self.secondary_metrics + self.dimension_fields:
            joinable_graphs = [j for j in field.join_graphs() if "merged_result" not in j]
            joinable = all(any(j in join_set for j in joinable_graphs) for join_set in joinable_sets)

            cannot_join = not joinable or len(joinable_sets) == 0
            measure_date_missing = field.canon_date not in canon_dates and field.field_type == "measure"
            if field.canon_date and (cannot_join or measure_date_missing):
                canon_dates.append(field.canon_date)
                joinable_sets.append(joinable_graphs)

        canon_dates = list(sorted(list(set(canon_dates))))
        for to_canon_date_name in canon_dates:
            for from_canon_date_name in canon_dates:
                if to_canon_date_name != from_canon_date_name:
                    from_canon_date = self.project.get_field_by_name(from_canon_date_name)
                    join_group_hash = self._join_hash_key(from_canon_date)
                    key = self._cte_name_from_parts(from_canon_date_name, join_group_hash)
                    canon_date_data = {"field": from_canon_date_name, "from_join_hash": key}
                    dimension_mapping[to_canon_date_name].append(canon_date_data)
                    join_hashes.append(key)

        sorted_joins = sorted(
            list(set(join_hashes)),
            key=lambda x: str(list(self.query_metrics.keys()).index(x)) if x in self.query_metrics else x,
        )
        return dimension_mapping, canon_dates, sorted_joins

    @staticmethod
    def _join_hash_key(field):
        # This makes the join hash out of all the joinable graphs, it's crucial to use this when
        # naming the CTEs so in subsequent checks we can match up all joinable fields,
        # not just fields in the same view
        joinable_graphs = [jg for jg in field.join_graphs() if "merged_result" not in jg]
        return "_".join(joinable_graphs)

    @staticmethod
    def _cte_name_from_parts(field_id: str, join_group_hash: str):
        canon_date = field_id.replace(".", "_")
        return f"{canon_date}__{join_group_hash}"

    @staticmethod
    def deduplicate_fields(field_dict: dict):
        # Get rid of duplicates while keeping order to make joining work properly
        return {k: sorted(list(set(v)), key=lambda x: v.index(x)) for k, v in field_dict.items()}

    @staticmethod
    def hash_dict(input_dict: dict):
        as_strings = {k: str(v) for k, v in input_dict.items()}
        return hashlib.sha256(json.dumps(as_strings, sort_keys=True).encode("utf-8")).hexdigest()
